"""
    Source file caching - load, save
"""
import logging
import re
import sys
import traceback
from bdb import BdbQuit
from contextlib import contextmanager
from csv import writer
from os import linesep
from pathlib import Path
from sys import exit
from time import time

import ezodf
import jsonpickle
import openpyxl
import xlrd
from netaddr import IPRange
from xlrd import XLRDError

from .config import Config, config_dir
from .dialogue import hit_any_key, is_yes
from .identifier import Identifier
from .parser import Parser
from .utils import lazy_print

logger = logging.getLogger(__name__)

__author__ = "Edvard Rejthar"
__date__ = "$Mar 23, 2015 8:33:24 PM$"

WHOIS_CACHE = ".convey-whois-cache.json"


def choose_file():
    print("Set path to the source log file.")
    print("? ", end="")

    # XX in the future, let's get rid of Tkinter. And don't impose it now if not really needed
    try:
        import tkinter as tk
        from tkinter.filedialog import askopenfilename
        root = tk.Tk()
        root.withdraw()  # show askopenfilename dialog without the Tkinter window
        file = askopenfilename()  # default is all file types
        print(file)
        return file
    except ImportError:
        print("Error importing Tkinter. Please specify the file name in the parameter or install `apt install python3-tk`.")
        return None


def read_stdin():
    if Config.get("daemon", get=bool):
        raise ConnectionRefusedError("STDIN missing")
    print("Write something to stdin. (End of transmission 2× <Ctrl>+d or <Enter>+<Ctrl>+d.)")
    return sys.stdin.read().rstrip().split("\n")  # rstrip \n at the end of the input


class Wrapper:
    def __init__(self, file_or_input, force_file=False, force_input=False,
                 types=None, fresh=False, reprocess=False, delete_cache=False):
        if delete_cache and Path(config_dir, WHOIS_CACHE).exists():
            Path(config_dir, WHOIS_CACHE).unlink()

        self.parser: Parser = None
        self.file = file = None
        self.stdin = stdin = None
        self.types = types
        self.whois_not_loaded = fresh
        self.last_hash = None
        try:
            case = int(Config.get("file_or_input"))
        except (ValueError, TypeError):
            case = 0

        if case == 5:
            force_file = True
        elif case == 6:
            force_input = True

        if force_input:
            stdin = file_or_input.split("\n") if file_or_input else read_stdin()
        elif force_file:
            file = file_or_input if file_or_input else choose_file()
        elif isinstance(file_or_input, Path) or (file_or_input and len(file_or_input) < 256 and Path(file_or_input).is_file()):
            # if longer than 255, it is most probably not a file but input
            file = file_or_input
        elif file_or_input:
            stdin = file_or_input.split("\n")
        else:  # choosing the file or input text
            if case == 0:
                case = 1 if is_yes("Do you want to input text (otherwise you'll be asked to choose a file name)?") else 2
            if case == 1:
                stdin = read_stdin()
            elif case == 2:
                file = choose_file()
            elif case == 3:
                stdin = read_stdin()
                if not stdin:
                    file = choose_file()
            elif case == 4:
                file = choose_file()
                if not file:
                    stdin = read_stdin()

        if not file and not stdin:
            print("No input. Program exit.")
            exit()
        elif stdin:
            Config.set_cache_dir(Path.cwd())
            self.cache_file = None
            self.stdin = stdin
            self.parser: Parser = Parser(stdin=stdin, types=self.types)
            return

        if not Path(file).is_file():
            print(f"File '{file}' not found.")
            exit()

        self.assure_cache_file(file)
        if self.cache_file.is_file() and not (fresh or reprocess):
            logger.info(f"File {self.file} has already been processed.")
            try:  # try to depickle
                self.parser = jsonpickle.decode(self.cache_file.read_text(), keys=True)
            except:
                print(traceback.format_exc())
                if not Config.error_caught():
                    input()
                else:
                    print("Cache file loading failed, let's process it all again. If you continue, cache gets deleted.")
                self.parser = None

        if self.parser:  # we have successfully loaded a parser
            if self.parser.source_file != self.file:  # file might have been moved to another location
                self.parser.source_file = self.file
            try:
                if self.parser.is_analyzed or self.parser.is_formatted:
                    self.parser.informer.sout_info()
                    logger.info("It seems the file has already been formatted.")
            except BdbQuit:  # we do not want to catch the exit signal from pdb
                print("Stopping.")
                exit()
            except Exception as e:
                print(e)
                print("Format of the file may have changed since last time. "
                      "Let's process it all again. If you continue, cache gets deleted.")
                if not Config.error_caught():
                    input()
                self.parser = None

        if not self.parser:
            self.clear()

        if not fresh:
            self.parser.ip_seen, self.parser.ranges = self.load_whois_cache()
            self.parser.refresh()
            self.parser.reset_whois(assure_init=True)
            # correction of a wrongly pickling: instead of {IPNetwork('...'): (IPNetwork('...'),
            # we see {IPNetwork('...'): (<jsonpickle.unpickler._IDProxy object at 0x...>,
            # Note that IPRange is pickled correctly.
            for prefix, o in self.parser.ranges.items():
                l = list(o)
                l[0] = prefix
                self.parser.ranges[prefix] = tuple(l)

    def assure_cache_file(self, file):
        self.file = Path(file).resolve()
        info = self.file.stat()
        hash_ = str(hash(info.st_size + round(info.st_mtime)))  # XX use st_size only?

        # cache-file with source file metadata
        Config.set_cache_dir(Path(self.file.parent, self.file.name + "_convey" + hash_))
        self.cache_file = Path(Config.get_cache_dir(), self.file.name + ".cache")

    def load_whois_cache(self):
        """ restore whois cache and remove expired results """
        p = Path(config_dir, WHOIS_CACHE)
        if Config.get("whois_cache", "FIELDS", get=bool) and p.exists():
            # XX if it's long, postpone via a thread that would block analysis
            event = lazy_print("... loading big WHOIS cache ...")
            ip_seen, ranges = jsonpickle.decode(p.read_text(), keys=True)
            ranges = {IPRange(k[0], k[1]): v for k, v in ranges.items()
                      if v[7] + Config.get("whois_ttl", "FIELDS", int) >= time()}
            nothing_to_save = True
            if Config.get("whois_delete_unknown", get=bool) and IPRange(0, 0) in ranges:
                # all IP addresses within an unknown prefix removed from cache
                del ranges[IPRange(0, 0)]
                nothing_to_save = False
            ip_seen = {k: v for k, v in ip_seen.items() if v in ranges}
            if nothing_to_save:  # count hash now so that we do not re-save whois cache if not changed while processing
                self._whois_changed(ranges, ip_seen)
            event.set()
            return ip_seen, ranges
        return {}, {}

    def _whois_changed(self, ranges, ip_seen):
        """ Quick hashing to identify if something changed to spare time.
              When cache json is 5.7 MB big, jsonpickling takes 3000 ms, hashing dict items 200 ms
              and hashing values only 100 ms."""
        logger.debug("Checking whois cache new records...")
        h = hash(frozenset(ranges.values())) + hash(frozenset(ip_seen.values()))
        if self.last_hash != h:
            self.last_hash = h
            return True

    ##
    # Store
    def save(self, last_chance=False):
        # chance to save the original file to the disk if reading from STDIN
        if not self.cache_file and \
                ((self.parser.stdout is not True and self.parser.stdout) or self.parser.is_formatted):
            # * cache_file does not exist = we have not written anything on the disk
            # * target_file exist = there is a destination to write (when splitting, no target_file specified)
            #       XX which may be changed because it is usual to preserve file
            # * stdout or is_formatted - there is something valuable to be saved
            target_file = self.parser.target_file or self.parser.source_file
            if not target_file:  # we were splitting so no target file exist
                target_file = self.parser.invent_file_str()
            if Config.get("output") is None:
                i = Config.get("save_stdin_output", get=int)
                save = False
                if i == 4 or (i == 3 and self.parser.is_analyzed):
                    save = True
                elif i == 2 or (i == 1 and self.parser.is_analyzed) and last_chance:
                    save = is_yes(f"Save to an output file {target_file}?")
            else:
                save = bool(Config.get("output"))
            if save:
                if not self.parser.target_file:
                    # if a split files were generated, we lost the connection because cache dir will change right now
                    # so we forget we've already split and make the file processable again (because split settings remained)
                    self.parser.is_split = False
                    self.parser.is_analyzed = False
                self.parser.source_file = target_file
                if self.parser.stdout is True:
                    # case: input has been processed and returned an empty result, ex: all rows filtered out
                    target_file.write_text("")
                else:  # XX I do not understand when we save self.parser.stdin
                    target_file.write_text(self.parser.stdout or linesep.join(self.parser.stdin))
                self.parser.stdin = None
                self.assure_cache_file(target_file)
            if self.parser.target_file:
                self.parser.saved_to_disk = bool(save)

        # serialize
        string = jsonpickle.encode(self.parser, keys=True)
        try:
            # Sometimes jsonpickling fails. But even if it does not fail, we are not sure everything is alright.
            # Once jsonpickle shuffled keys and values of self.parser.ip_seen dict (when this dict was yet part of self.parser).
            # There must be problem with a bad reference.
            # – once I encountered "get[7] datetime" instead of "get[0] prefix" as the value.
            # I narrowed the case to a CSV of 3 lines: any IPs to whom I added a netname col. When decoded, values were shuffled.
            # Unfortunately, I have not been able to simulate this behaviour with another simpler object than self.parser
            # so that I could raise an official issue.
            jsonpickle.decode(string, keys=True)
        except Exception:
            hit_any_key("The program state is not picklable by 'jsonpickle' module. "
                  "Continuing will provide a file that will have to be reanalysed. "
                  "You may post this as a bug to the project issue tracker.")
            Config.error_caught()

        # save cache file
        if self.cache_file:  # cache_file does not exist typically if reading from STDIN
            if self.parser.ranges and Config.get("whois_cache", "FIELDS", get=bool):
                # we extract whois info from self.parser and save it apart for every convey instance
                if self.whois_not_loaded:  # if we wanted a fresh result, global whois cache was not used and we have to merge it
                    ip_seen, ranges = self.load_whois_cache()
                    ip_seen = {**ip_seen, **self.parser.ip_seen}
                    ranges = {**ranges, **self.parser.ranges}
                    self.whois_not_loaded = False
                else:
                    ip_seen, ranges = self.parser.ip_seen, self.parser.ranges
                if self._whois_changed(ranges, ip_seen):
                    # note that ip_seen MUST be placed before ranges due to https://github.com/jsonpickle/jsonpickle/issues/280
                    # That way, a netaddr object (IPNetwork, IPRange) are defined as value in ip_seen and not as key in range.
                    # Update version 1.3.1: However this was not enough, serializing object as dict keys was still a problem.
                    # So we are manually converting them to int-tuples.
                    event = lazy_print("... saving big WHOIS cache ...")
                    ranges_serializable = {(k.first, k.last): v for k, v in ranges.items()}
                    encoded = jsonpickle.encode([ip_seen, ranges_serializable], keys=True)
                    # noinspection PyBroadException
                    try:
                        jsonpickle.decode(encoded, keys=True)
                    except Exception:  # again, I met a strangely formed JSON
                        type_, value, tb = sys.exc_info()
                        body = f"```bash\n{traceback.format_exc()}```\n\n" \
                               f"```json5\n{tb.tb_next.tb_frame.f_locals}\n```\n\n" \
                               f"```json5\n{ip_seen}```\n\n```json5\n{ranges}```"
                        print("The program will recover but without saving WHOIS cache.")
                        Config.github_issue("Cannot jsonpickle whois", body)
                    else:
                        Path(config_dir, WHOIS_CACHE).write_text(encoded)
                    finally:
                        event.set()
            self.cache_file.write_text(string)  # save cache

    def clear(self):
        self.check_ods() or self.check_xlsx() or self.check_xls() or self.check_log()

        self.parser = Parser(self.file, self.stdin, self.types)
        self.save()

    def check_log(self):
        """ Check if the contents is a CSV and not just a log
         ex: "06:25:13.378767 IP 142.234.39.36.51354 > 195.250.148.86.80: Flags [S], seq 1852455482, win 29200, length 0"
         """
        re_ip_with_port = re.compile("((\d{1,3}\.){4})(\d+)")
        re_log_line = re.compile(r"([^\s]*)\sIP\s([^\s]*)\s>\s([^\s:]*)")
        _, sample, _ = Identifier(None).get_sample(self.file)
        if sample and re_log_line.match(sample[0]):

            if is_yes("This seems like a log file. Do you wish to transform it to CSV first?"):
                with self.rework() as target:
                    target.writerow(["time", "source", "src_port", "dst", "dst_port"])
                    with open(self.file) as f:
                        for line in f.readlines():
                            try:
                                res = re_log_line.search(line).groups()
                            except AttributeError:
                                print("Error", line)
                                break
                            else:
                                def parse_ip(val):
                                    m = re_ip_with_port.match(val)
                                    if m:
                                        return m[1][:-1], m[3]
                                    else:
                                        return "", ""

                                timestamp = res[0]
                                source, src_port = parse_ip(res[1])
                                dst, dst_port = parse_ip(res[2])
                                target.writerow([timestamp, source, src_port, dst, dst_port])
            return True

    def check_ods(self):
        """ Check if the contents is an ODS file """
        try:
            doc = ezodf.opendoc(self.file)
        except KeyError:
            pass
        else:
            if is_yes("This seems like an ODS file. Do you wish to transform it to CSV first?"):
                sheet = doc.sheets[0]
                with self.rework() as target:
                    for i, row in enumerate(sheet.rows()):
                        r = []
                        for j, cell in enumerate(row):
                            r.append(cell.value)
                        target.writerow(r)
            return True

    def check_xls(self):
        """ Check if the contents is a XLS file """
        try:
            wb = xlrd.open_workbook(self.file)
            sh = wb.sheets()[0]
        except (XLRDError, IndexError):
            pass
        else:
            if is_yes("This seems like a XLS file. Do you wish to transform it to CSV first?"):
                with self.rework() as target:
                    for row in range(sh.nrows):
                        target.writerow(sh.row_values(row))
            return True

    def check_xlsx(self):
        """ Check if the contents is a XLSX file """
        try:
            wb = openpyxl.load_workbook(self.file, read_only=True, keep_vba=False, data_only=True, keep_links=False)
            sh = wb[wb.sheetnames[0]]
        except (openpyxl.utils.exceptions.InvalidFileException, IndexError):
            pass
        else:
            if is_yes("This seems like a XLSX file. Do you wish to transform it to CSV first?"):
                with self.rework() as target:
                    for row in sh.values:
                        target.writerow(row)
            return True


    @contextmanager
    def rework(self):
        target = Path(str(self.file) + ".csv")
        if target.exists():
            if is_yes(f"It seems the file {target.absolute()} already exists. Do you wish to process it?"):
                try:
                    yield None
                except AttributeError:
                    pass
                finally:
                    # XX however, this resets the cache of the file because we continue in the clear() method
                    #   We rather do want to directly go to the file.
                    self.assure_cache_file(target.absolute())  # changes self.file
                    return
            else:
                print(f"The file {target} already exist, cannot be recreated.")
                exit()
        # noinspection PyBroadException
        try:
            with target.open("w") as f:
                yield writer(f)
        except Exception as e:
            print(f"Could not convert.")
            logger.error(e)
            Config.error_caught()
            raise
        else:
            hit_any_key(f"Successfully written to {target.absolute()}.")
            # XX remove old cache directory if empty
            self.assure_cache_file(target.absolute())  # changes self.file



import logging
import re
from sre_constants import error

from prompt_toolkit import PromptSession, ANSI, HTML
from prompt_toolkit.application import get_app
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.shortcuts import clear
from prompt_toolkit.styles import Style
from prompt_toolkit.styles import merge_styles
from prompt_toolkit.styles.pygments import style_from_pygments_cls
from prompt_toolkit.validation import Validator, ValidationError
from pygments.lexer import RegexLexer, include
from pygments.lexers.python import Python3Lexer
from pygments.styles import get_style_by_name
from pygments.token import Token
from tabulate import tabulate

from convey import PickInput
from .config import Config, console_handler
from .dialogue import Cancelled
from .types import Type, Types


def yellow_no_end(s):
    return f"\033[0;41m{s}"


def yellow(s, error=False):
    s = f"\033[0;33m{s}\033[0m"
    if error:
        s += red("!")
    return s


def blue(s):
    return f"\033[0;36m{s}\033[0m"


def red(s):
    return f"\033[0;41m{s}\033[0m"


class RegularLexer(RegexLexer):
    tokens = {
        'root': [
            (r'\\^', Token.START_TEXT),
            (r'\$', Token.END_TEXT),
            (r'(\[\^)', Token.SQUARE_NEGATIVE, 'square-negative'),
            (r'(\[)', Token.SQUARE, 'square'),
            (r'\(', Token.ROUND, 'round'),
            (r'\{', Token.CURLY, 'curly'),
            include("grammar")
        ],
        'grammar': [
            (r'\\w', Token.W),
            (r'\\d', Token.D),
            (r'\.', Token.DOT),
            (r'\*', Token.ASTERISK),
            (r'\+', Token.PLUS),
            (r'\?', Token.QUESTION),
            (r'\//', Token.SLASH),
            (r'\|', Token.PIPE),
            (r'.', Token.TEXT),
        ],
        'round': [
            (r'\)', Token.ROUND, '#pop'),
            include("grammar"),
        ],
        'curly': [
            (r'\}', Token.CURLY, '#pop'),
            (r'.', Token.CURLY_TEXT),
        ],
        'square': [
            (r'\]', Token.SQUARE, '#pop'),
            (r'.', Token.SQUARE_TEXT),
        ],
        'square-negative': [
            (r'\]', Token.SQUARE_NEGATIVE, '#pop'),
            (r'.', Token.SQUARE_TEXT_NEGATIVE),
        ]
    }


# test_regex = "\w\d.*+?//| (abc)  [abc] [^abc] {abc}"

# reg_style = style_from_pygments_cls(get_style_by_name('colorful'))
reg_style = style_from_pygments_cls(get_style_by_name('friendly'))
reg_style = merge_styles([reg_style, Style.from_dict({
    'pygments.start_text': '#F00',
    'pygments.end_text': '#F00',
    'pygments.text': '#fff bg:#005',
    'pygments.w': '#000 bg:#00ffd7',
    'pygments.d': '#000 bg:#005fff',
    'pygments.dot': '#000 bg:#ffff00',
    'pygments.asterisk': '#000 bg:#ff5522',
    'pygments.plus': '#000 bg:#ff7755',
    'pygments.question': '#000 bg:#ff9977',
    'pygments.slash': '#f00 bg:#000',
    'pygments.pipe': '#0F0',
    'pygments.round': '#0F0',
    'pygments.curly': '#000 bg:#ffff55',
    'pygments.curly_text': '#000 bg:#ffff55',
    'pygments.square': '#fff bg:#0000ff',
    'pygments.square_text': '#fff bg:#0000ff',
    'pygments.square_negative': '#fff bg:#870000',
    'pygments.square_text_negative': '#fff bg:#870000',
})])
bottom_plain_style = Style.from_dict({
    'bottom-toolbar': 'noreverse',
})


def _code_method(e, x):
    """ simulate real code execution in the Identifier """
    l = locals()
    try:
        exec(compile(e, '', 'exec'), l)
    except:
        return ""
    x = l["x"]
    return x


def _reg_method(line, search, replace=None):
    """ simulate real reg ex matching in the Identifier """
    match = search.search(line)
    reg_s = ""

    if not match:
        return "", "", "", blue(line)
    groups = match.groups()
    if not groups:
        groups_preview = [match.group(0)]
    else:
        groups_preview = ["{0}: " + str(match.group(0))] + ["{" + str(i + 1) + "}: " + g for i, g in enumerate(groups)]

    if not replace:
        if not groups:
            reg_m = match.group(0)
        else:
            reg_m = match.group(1)
        reg_s = search.sub("", line)
    else:
        try:
            reg_m = replace.format(match.group(0), *[g for g in groups])
            # we convert "str{0}" → "\g<0>" (works better than conversion to a mere "\0" that may result to ambiguity
            try:
                replace_pattern = re.sub(r"{(\d+)}", r"\\g<\1>", replace)
                reg_s = search.sub(replace_pattern, line)
            except error:
                reg_s = "!"
        except IndexError:
            reg_m = ""
    span = match.span(0)
    return groups_preview, reg_m, reg_s, blue(line[0:span[0]]) + red(line[slice(*span)]) + blue(line[span[1]:])


class ReValidator(Validator):
    def validate(self, document):
        text = document.text
        try:
            re.compile(text)
        except re.error:
            for i in range(len(text)):
                try:
                    text = text[:-1]
                except re.error:
                    pass
                else:
                    break
            raise ValidationError(message='Cannot be used as a regexp.', cursor_position=len(text) - 1)


class TypeValidator(Validator):
    def validate(self, document):
        if document.text not in ["reg_m", "reg_s"]:
            raise ValidationError(message='Type reg_m or reg_s')


class Preview:

    def __init__(self, source_field, source_type: "Type", target_type: "Type" = None):
        self.source_field = source_field
        self.source_type = source_type
        self.target_type = target_type
        # common part of every preview
        self.samples = source_field.get_samples(supposed_type=source_type, target_type=target_type)

        # define key self.bindings
        self.session = self.reset_session()
        self.bindings = KeyBindings()

        # cancel on control-c (escape inaccessible, waits for another keystroke)
        @self.bindings.add('c-c')
        def _(_):
            get_app().exit(False)

        # exit on control+d (because the default exit command alt-enter seems unintuitive) (control+enter unacessible)
        @self.bindings.add('c-d')
        def _(_):
            get_app().exit(self.session.layout.current_buffer.text)

        # define styling
        self.style = merge_styles([reg_style, bottom_plain_style])

        # Init variables that may be used in methods
        self.get_toolbar_row = self.search = self.replace = self.phase = self.chosen_type =  None

    def standard_toolbar(self):
        """ define bottom preview toolbar """
        rows = []
        level = Config.verbosity  # temporarily suppress info messages like 'NMAPing...' that would break up the wizzard layout
        console_handler.setLevel(logging.WARNING)
        for line in self.samples:
            val = self.get_toolbar_row(self.session.layout.current_buffer.text, line)
            rows.append((f"\033[0;36m{line}\033[0m", f"\033[0;33m{val}\033[0m"))  # blue and yellow
        console_handler.setLevel(level)

        return ANSI('\nPreview\n' + tabulate(rows, headers=("original", "result"), tablefmt="github"))

    def code(self):
        """ code preview specific part """

        def get_toolbar_row_code(text, line):
            return _code_method(text, line)

        self.get_toolbar_row = get_toolbar_row_code

        def get_prompt():
            s = '<b>Ctrl+D</b>' if "\n" in self.session.layout.current_buffer.text else 'Ctrl+D'
            return HTML(f'Code <i>({s} to confirm)</i>: ')

        # prints the application
        text = self.reset_session().prompt(get_prompt, bottom_toolbar=self.standard_toolbar, style=self.style,
                                           lexer=PygmentsLexer(Python3Lexer), multiline=True, key_bindings=self.bindings)
        return text

    def pick_input(self, o: PickInput):
        """ code preview specific part """

        def get_toolbar_row_pick_input(text, line):
            try:
                val = o.get_lambda(text)(line)
            except Exception as e:
                val = str(e)
            return val

        self.get_toolbar_row = get_toolbar_row_pick_input

        def get_prompt():
            return HTML(o.description + ": ")

        # prints the application
        text = self.reset_session().prompt(get_prompt, bottom_toolbar=self.standard_toolbar, style=self.style,
                                           default=o.default or "", lexer=PygmentsLexer(Python3Lexer), key_bindings=self.bindings)
        if text is False:
            raise Cancelled
        return text

    def reg(self):
        """ regex preview specific part """
        # self.ask_search = True
        self.phase = None
        self.search = ""
        self.replace = ""
        self.target_type = [self.target_type] if self.target_type != Types.reg else [Types.reg_m, Types.reg_s]
        self.chosen_type = None

        # @self.bindings.add('escape', 'left')  # alt-left
        @self.bindings.add('c-i')  # tab to jump back
        def _(_):
            if self.phase > 1:
                clear()
                self.phase = "continue"
            self.session.app.exit(self.session.layout.current_buffer.text)

        if len(self.target_type) > 1:
            def toggle_type(default):
                if self.chosen_type is None:
                    self.chosen_type = default
                else:
                    self.chosen_type = Types.reg_m if self.chosen_type == Types.reg_s else Types.reg_s
                if self.phase == 3:
                    self.session.layout.current_buffer.text = str(self.chosen_type)

            @self.bindings.add('escape', 'left')  # cycle preferred method
            def _(_):
                toggle_type(Types.reg_s)

            @self.bindings.add('escape', 'right')  # cycle preferred method
            def _(_):
                toggle_type(Types.reg_m)

        # prints the application
        options = {"bottom_toolbar": self.reg_toolbar, "style": self.style,
                   "lexer": PygmentsLexer(RegularLexer), "multiline": False, "key_bindings": self.bindings}
        while True:
            # self.ask_search = False
            self.phase = 1
            self.search = self.reset_session().prompt('Regular match: ', **options, default=self.search,
                                                      validator=ReValidator())
            if self.search is False:
                raise Cancelled
            if self.phase == "continue":
                continue
            self.phase = 2
            self.replace = self.reset_session().prompt('Regular replace: ', **options, default=self.replace,
                                                       rprompt=HTML('hit <b>tab</b> to jump back'))
            if self.replace is False:
                raise Cancelled
            if self.phase == "continue":
                continue
            self.phase = 3
            if len(self.target_type) > 1:  # we have to choose the column
                if self.chosen_type is None:
                    # we are not using groups, match has no sense
                    self.chosen_type = Types.reg_s if "{" not in self.replace and self.replace else Types.reg_m
                type_ = self.reset_session().prompt('Do you prefer match or substitution? ', **options,
                                                    default=str(self.chosen_type),
                                                    rprompt=HTML('choose with <b>Alt+←/→</b>'), validator=TypeValidator())
                if self.phase == "continue":
                    continue
                if type_:
                    self.target_type = Types.reg_s if type_ == "reg_s" else Types.reg_m
                elif type_ is False:
                    raise Cancelled
                else:
                    self.target_type = self.chosen_type
            else:
                self.target_type = self.target_type[0]
            break
        return self.search, self.replace, self.target_type

    def reset_session(self):
        """ Calling
            self.session = PromptSession()
            self.session.prompt(validator=V)
            self.session.prompt(validator=None) → would still have previous validator

        """
        self.session = PromptSession()
        return self.session

    # define bottom preview toolbar
    def reg_toolbar(self):
        helper = ""
        rows = []
        match = []
        for line in self.samples:
            text = self.session.layout.current_buffer.text
            error = False  # prompt text is erroneous, probably not completed yet
            while True:
                row = []
                if self.phase == 1:  # first phase - searching for match string
                    contents = text
                    try:
                        search_re = re.compile(contents)
                    except re.error:
                        text = text[:-1]
                        error = True
                        continue
                    match, reg_m_preview, reg_s_preview, line = _reg_method(line, search_re, self.replace)
                    row.append("\n".join([yellow(m) for m in match]))  # colorize lines separately
                    if error:
                        row[-1] += red("!")
                    reg_m_preview = blue(self.highlight(reg_m_preview, Types.reg_m))
                    reg_s_preview = blue(self.highlight(reg_s_preview, Types.reg_s))
                    helper = r". - any char, \w - word, \d - digit, () - matching group"
                else:  # second phase - searching for replace string (and third phase just diplaying)
                    contents = self.search
                    search_re = re.compile(contents)
                    try:
                        match, reg_m_preview, reg_s_preview, line = _reg_method(line, search_re,
                                                                                text if self.phase == 2 else self.replace)
                    except ValueError:
                        text = text[:-1]
                        error = True
                        continue
                    # blue, blue, yellow
                    row.append("\n".join([blue(m) for m in match]))
                    reg_m_preview = yellow(self.highlight(reg_m_preview, Types.reg_m), error)
                    reg_s_preview = yellow(self.highlight(reg_s_preview, Types.reg_s), error)
                    helper = "Access matched parts with {0}, ex: string {1} string."
                if Types.reg_m in self.target_type:
                    row.append(reg_m_preview)
                if Types.reg_s in self.target_type:
                    if contents:
                        row.append(reg_s_preview)
                    else:
                        row.append("")
                row.insert(0, line)
                rows.append(row)
                break
        headers = ["original", "groups" if len(match) > 1 else "group {0}"]
        if Types.reg_m in self.target_type:
            headers.append(self.highlight("match", Types.reg_m))
        if Types.reg_s in self.target_type:
            headers.append(self.highlight("substitution", Types.reg_s))
            # 'grid' handles multiline rows well, 'github' handles them bad, despite the documentation
        return ANSI('\n\nPreview\n' + tabulate(rows, headers=headers, tablefmt="grid") + "\n\n" + helper)

    def highlight(self, s, current):
        if self.chosen_type is None:
            return s
        elif self.chosen_type == current:
            return f"\033[1m{s}\033[0m"
        else:
            return f"\033[38;5;7m{s}\033[0m"



from __future__ import annotations
import csv
from dataclasses import dataclass
import logging
from typing import TYPE_CHECKING, Callable, List, Optional, Type

from attrs import define, asdict, field
from difflib import SequenceMatcher

if TYPE_CHECKING:
    from .field import Field
    from .parser import Parser

logger = logging.getLogger(__name__)


@dataclass
class BareField:
    """ This will become a Field once the parser is ready.
    Note: This cannot be an attrs class because it would be converted to dict by FlagController.read too early.
    """
    task: str


def Column(): return field(converter=lambda x: BareField(x) if x is not None else None, default=None)
def Path(): return field(default=None)


class FlagController:
    def __init__(self, parser: Parser):
        self.parser = parser

    def read(self, flag_type: Type[Flag], val: str):
        """ Sometimes ",".split is not enough, they can use quotes and commas in our mighty CLI.
        Besides, we convert user's will to specific Field.
        Ex: `--merge gif.csv,2` ->  MergeFlag(gif.csv, 2) -> {remote_path: gif.csv, remote_col_i: Field(col_i=2)}
        """
        flag = asdict(flag_type(*next(csv.reader([val]))))

        # post processing
        for k, v in flag.items():
            if isinstance(v, BareField):  # convert BareField to Field, we have the parser now
                flag[k] = self.parser.identifier.get_column_i(v.task, check=True)
        return flag


# XX we should convert more flags to adapt Flag
class Flag:
    """ Flag convert CLI input to an Action """
    pass


@define
class MergeFlag(Flag):
    remote_path = Path()
    remote_col_i = Column()
    local_col_i = Column()



from pathlib import Path

from setuptools import setup

from convey import __version__

# using the same libraries in requirements.txt because after many articles I didn't understand any good reason why I shouldn't
requirements = ""
p = Path("requirements.txt")
if p.exists():
    requirements = p.read_text()

# load long description
p = Path("README.md")
if p.exists():
    long_description = p.read_text()

setup(
    name='convey',
    version=__version__,
    packages=['convey'],
    author='Edvard Rejthar',
    author_email='edvard.rejthar@nic.cz',
    url='https://github.com/CZ-NIC/convey',
    license='GNU GPLv3',
    description='CSV processing and mutual conversion of web related data types',
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[requirements.split("\n")],
    entry_points={
        'console_scripts': [
            'convey = convey.__main__:main',
        ],
    },
    include_package_data=True
)



from __future__ import annotations
import subprocess
import sys
from collections import deque
from csv import writer
from datetime import datetime
from io import StringIO
from itertools import cycle, count
from math import ceil, log10
from pathlib import Path
from threading import Event, Thread
from time import sleep
from typing import TYPE_CHECKING

import humanize
from colorama import Fore
from tabulate import tabulate

from .config import Config, get_terminal_size
from .types import Aggregate
from .whois import Whois

if TYPE_CHECKING:
    from .parser import Parser


class Informer:
    """ Prints analysis data in nice manner. """

    def __init__(self, parser: Parser):
        self.parser = parser
        self.queue = deque(maxlen=10)
        self.stdout = sys.stdout
        self.stats_stop = Event()
        "if ._flag=True → ends, ._flag=1 → pauses, ._flag=False → runs"

    def sout_info(self, clear=True, full=False):
        """ Prints file information on the display. """
        if Config.is_quiet():
            return

        stdout = StringIO()

        if clear and not Config.get("daemon", get=bool):
            stdout.write("\x1b[2J\x1b[H")

        def print_s(s):
            stdout.write(s + "\n")

        l = []
        p = self.parser
        se = p.settings
        l.append("Reading STDIN" if p.stdin else "Source file: " + str(p.source_file))
        if p.dialect:
            s = p.dialect.delimiter.replace("\t", "TAB")
            if se["dialect"] and p.dialect.delimiter != se["dialect"].delimiter:
                s2 = se['dialect'].delimiter.replace("\t", "TAB")
                s = f"{s} → {s2}"
            l.append(f"delimiter: '{Fore.YELLOW}{s}{Fore.RESET}'")

            s = p.dialect.quotechar
            if se["dialect"] and p.dialect.quotechar != se["dialect"].quotechar:
                s = f"{s} → {se['dialect'].quotechar}"
            l.append(f"quoting: '{Fore.YELLOW}{s}{Fore.RESET}'")
        if p.has_header is not None:
            l.append("header: " + (("remove" if se["header"] is False else "used") if p.has_header else "not used"))
        if se["filter"]:
            l.append("Filter: " + ", ".join([f"{p.fields[f].name} {'' if include else '!'}= {val}"
                                             for include, f, val in se["filter"]]))
        if se["unique"]:
            l.append("Unique col: " + ", ".join([p.fields[f].name for f in se["unique"]]))
        if se["split"] or se["split"] == 0:
            l.append("Split by: {}".format(p.fields[se["split"]]))
        if se["aggregate"]:
            v = ", ".join(f"{fn.__name__}({col.name})" for fn, col in se["aggregate"].actions)
            if se["aggregate"].group_by is not None:
                l.append(f"Group by {se['aggregate'].group_by}: " + v)
            else:
                l.append("Aggregate: " + v)

        # XX
        # if self.parser.redo_invalids is not None:
        #    l.append("Redo invalids: " + str(self.parser.redo_invalids))
        stdout.write(", ".join(l))

        l3 = []
        for f in p.fields:
            if not f.is_new and f.has_clear_type():
                t = ", ".join([str(t) for t in f.possible_types])
                l3.append(f.color(f"{f} ({t})"))
        stdout.write("\nIdentified columns: " + ", ".join(l3))
        if se["add"] or se["merge"]:
            l2 = []
            l2.extend(f.color(f"{f} (from {str(f.source_field)})") for f in se["add"])
            l2.extend(f"{ma.remote_parser.source_file} (from {str(ma.local_column)})" for ma in se["merge"])
            stdout.write("\nComputed columns: " + ", ".join(l2))
        l = []
        progress = 0
        if p.line_count:
            if p.ip_count:
                stdout.write(", {} IPs".format(p.ip_count))
            elif p.ip_count_guess:
                stdout.write(", around {} IPs".format(p.ip_count_guess))
            l.append("Log lines processed: {}/{}, {} %".format(p.line_count, p.lines_total,
                                                               ceil(100 * p.line_count / p.lines_total)))
            progress = p.line_count / p.lines_total
        else:
            l.append("Log lines: {}".format(p.lines_total))
        if p.time_end:
            l.append("{}".format(p.time_end - p.time_start))
        elif p.time_start:
            l.append(f"{datetime.now().replace(microsecond=0) - p.time_start}")
            l.append(f"{p.velocity} lines / s")
            l.append(f"{p.processor.descriptors_count} file descriptors open")
        if p.queued_lines_count and hasattr(Whois, "queued_ips"):
            # XX it would be cleaner if the case that Whois.queued_ips (and Whois.quota) does not exist
            #    will not happen; but it will when queued, interrupted and launched again.
            if Whois.queued_ips and len(Whois.queued_ips) != p.queued_lines_count:
                # why Whois.queued_ips > 0: Quota has ended and set of queued IPs has been emptied
                # ... but the lines are still queued. We lost the information about the number of unique queued IPs.
                v = f" lines ({len(Whois.queued_ips)} unique IPs)"
            else:
                v = " lines"
            s = f"skipped {p.queued_lines_count}{v} due to LACNIC quota"
            if Whois.quota.is_running():
                s += f" till {Whois.quota.time()}"
            l.append(s)
        r = ", ".join(l)
        if progress:
            _, width = get_terminal_size()
            progress = round(progress * width) if width else 0
            if len(r) < progress:
                r += " " * (progress - len(r))
            r = f"\033[7m{r[:progress]}\033[0m" + r[progress:]
        stdout.write("\n" + r + "\n")
        if p.whois_stats:
            print_s("Whois servers asked: " + ", ".join(key + " (" + str(val) + "×)" for key, val in p.whois_stats.items())
                    + f"; {len(p.ranges)} prefixes discovered")

        print_s("\nSample:\n" + "".join(p.sample[:4]))  # show first 3rd lines

        if p.is_formatted:  # show how would the result be alike
            full_rows, rows = p.get_sample_values()

            first_line_length = tabulate(rows, headers=[f.get(True, color=False) for f in p.fields]).split("\n")[0]
            if rows and len(first_line_length) <= get_terminal_size()[1]:
                # print big and nice table because we do not care about the dialect and terminal is wide enough
                # we do not display dialect change in the big preview
                print_s("\033[0;36mPreview:\033[0m")
                header = [f.get(True, line_chosen=se["header"] is not False) for f in p.fields]
                print_s(tabulate(rows, headers=header))
            else:
                # print the rows in the same way so that they optically match the Sample above
                print_s("\033[0;36mCompact preview:\033[0m")
                cw = writer(stdout, dialect=se["dialect"] or p.dialect)
                if se["header"] is not False and p.has_header:
                    cw.writerow([f.get() for f in p.fields])
                for r in full_rows:
                    cw.writerow(r)

        output = Config.get("output")
        if output:
            print_s(f"Output file specified: {output}")

        if p.aggregation:  # an aggregation has finished
            print_s("\n")
            if len(p.aggregation) == 1:
                print_s(self.get_aggregation(next(iter(p.aggregation.values())), color=True, limit=8))
            else:
                print_s("Aggregating in split files...")

        if p.is_analyzed:
            if p.saved_to_disk is False:
                print_s("\n** Processing completed, results were not saved to a file yet.")
                print_s(tabulate(p.stdout_sample, headers="firstrow" if p.has_header else ()))
            elif p.saved_to_disk:
                print_s(f"\n** Processing completed: Result file in {p.target_file}")
            else:
                abroad, local, non_deliverable, totals = map(p.stats.get, (
                    'abroad', 'local', 'non_deliverable', 'totals'))

                print_s(f"** Processing completed: {totals} result files in {Config.get_cache_dir()}")
                # local = Contacts.count_mails(self.attachments.keys(), abusemails_only=True)
                # abroad = Contacts.count_mails(self.attachments.keys(), abroads_only=True)
                if totals == non_deliverable:
                    print_s("* It seems no file is meant to serve as an e-mail attachment.")
                else:
                    if local[0] + abroad[0] == 0:
                        print_s(f"Already sent all {abroad[1]} abroad e-mails and {local[1]} local e-mails")
                    elif local[1] + abroad[1] > 1:
                        print_s(f"Already sent {abroad[1]}/{sum(abroad)} abroad e-mails and {local[1]}/{sum(local)} local e-mails")
                    else:
                        print_s(f"* {abroad[0]} files seem to be attachments for abroad e-mails\n* {local[0]} for local e-mails")
                    if non_deliverable:
                        print_s(f"* {non_deliverable} files undeliverable")

                if Config.get('testing'):
                    print_s(
                        "\n*** TESTING MOD - mails will be send to mail {} ***\n (For turning off testing mode set `testing = False` in config.ini.)".format(
                            Config.get('testing_mail')))

            stat = self.get_stats_phrase()
            if stat:
                print_s("\n Statistics overview:\n" + stat)
            """
            XX
            if parser.abuseReg.stat("records", False):
                print_s("Couldn't find {} abusemails for {}× IP.".format(parser.reg["local"].stat("records", False), parser.reg["local"].stat("ips", False)))
            if parser.countryReg.stat("records", False):
                print_s("Couldn't find {} csirtmails for {}× IP.".format(parser.reg["abroad"].stat("records", False), parser.reg["abroad"].stat("ips", False)))
            """

        if full:
            # XX subprocess.run("less", input=v, encoding="utf-8")
            # XX show aggregation too
            if len(p.ranges.items()):
                rows = []
                for prefix, o in p.ranges.items():
                    prefix, location, incident, asn, netname, country, abusemail, timestamp = o
                    rows.append((prefix, location, incident or "-", asn or "-", netname or "-"))
                print_s("\n\n** Whois information overview **\n" +
                        tabulate(rows, headers=("prefix", "location", "contact", "asn", "netname")))
            else:
                print_s("No whois information available.")

            if p.is_split:
                rows = []
                for o in p.attachments:
                    rows.append((o.filename,
                                 {True: "abroad", False: "✓", None: "×"}[o.abroad],
                                 {True: "✓", False: "error", None: "no"}[o.sent],
                                 humanize.naturalsize(o.path.stat().st_size),
                                 ))
                print_s("\n\n** Generated files overview **\n" + tabulate(rows, headers=("file", "deliverable", "sent", "size")))
            # else:
            #     print_s("Files overview not needed – everything have been processed into a single file.")

            print_s("\n\nPress Enter to continue...")

        # atomic print out
        self.stdout.write(stdout.getvalue())
        if self.queue:  # what has been printed out during processing stays on the screen
            self.stdout.write("".join(self.queue))
            # XX if random true, pops out an element so that it wont stay forever

    def get_aggregation(self, data, color=False, limit=None, nice=True):
        """

        @param data:
        @param color:
        @param limit:
        @param nice: If true, tabulated result returned, else we get (header, rows) tuple.
        @return:
        """
        form = lambda v, fmt: f"\033[{fmt}m{v}\033[0m" if color else v
        header = []
        grouping = self.parser.settings["aggregate"].group_by is not None
        if grouping:
            header.append(form(self.parser.settings["aggregate"].group_by.name, 36))
        header.extend([form(f"{fn.__name__}({col})", 33) for fn, col in self.parser.settings["aggregate"].actions])

        rows = []
        generators = cycle(g[0] for g in self.parser.settings["aggregate"].actions)
        # sorting by first column (may be a bottleneck, re-sorting every time)
        dd = sorted(data.items(), key=lambda x: x[1][0][1], reverse=True)
        for i, (grouped_el, d) in enumerate(dd):
            if limit and i == limit:
                rows.append(["..." for fn, count in d])
                if grouping:
                    rows[-1].insert(0, "...")
                break
            if grouped_el is None and len(self.parser.settings["aggregate"].actions) == 1 \
                    and next(generators) == Aggregate.list:
                # This is the total row
                # We are aggregating only single thing which is list.
                # That list would comprehend all of the values in the column. We omit it.
                continue

            # rows.append([form(count if fn.__name__ in ("count", "list") else round(count * 100) / 100, 33) for fn, count in d])
            rows.append([form(count if next(generators) not in (Aggregate.sum, Aggregate.avg) else round(count, 2), 33) for
                         fn, count in d])
            if grouping:
                rows[-1].insert(0, form("total" if grouped_el is None else grouped_el, 36))
        # floatfmt - display numbers longers than 15 as the scientific 1e+15, not numbers bigger than a million only
        return tabulate(rows, header, floatfmt=".15g") if nice else (header, rows)

    def get_stats_phrase(self, generate=False):
        """ Prints phrase "Totally {} of unique IPs in {} countries...": """
        st = self.parser.stat

        ip_unique = st("ip_unique")
        abusemail_local = st("abusemail_local")
        abusemail_abroad = st("abusemail_abroad")
        abusemail_unofficial = st("abusemail_unofficial")  # subset of abusemail_abroad
        ip_local_unknown = st("ip_local_unknown")
        ip_local_known = st("ip_local_known")
        ip_abroad_known = st("ip_abroad_known")
        ip_abroad_unknown = st("ip_abroad_unknown")
        prefix_local_known = st("prefix_local_known")
        prefix_local_unknown = st("prefix_local_unknown")
        prefix_abroad_known = st("prefix_abroad_known")
        prefix_abroad_unknown = st("prefix_abroad_unknown")
        prefix_csirtmail_unofficial = st("prefix_csirtmail_unofficial")
        ip_csirtmail_unofficial = st("ip_csirtmail_unofficial")
        ip_csirtmail_known = st("ip_csirtmail_known")
        csirtmail_unofficial = st("csirtmail_unofficial")
        csirtmail_known = st("csirtmail_known")

        """         XX
        invalidLines = self.parser.invalidReg.stat()
        """

        res = []
        if ip_unique > 0:
            res.append(f"Totally {ip_unique} of unique IPs")

        # * deliverable to 67 national/governmental CSIRTs (116490 IPs),
        #       no official contact for 554 abroad e-mail addreses of 107 countries (1615 IP in 45 prefixes)
        # * 445 e-mail addresses (38454 IP in 784 prefixes, unknown contact for 84 IP in 5 prefixes)
        # * 6848 abroad e-mail addresses (34515 IP in 84 prefixes, unknown contact for 84 in 4 prefixes)

        # interesting if using incident-contact field
        if ip_csirtmail_known or ip_csirtmail_unofficial:
            s = s2 = ""
            if ip_csirtmail_known:
                s = f"({ip_csirtmail_known} IP)"
            if ip_csirtmail_unofficial:
                s2 = f", no official contact for {abusemail_unofficial} abroad e-mail addresses" \
                     f" in {csirtmail_unofficial} countries ({ip_csirtmail_unofficial} IP in {prefix_csirtmail_unofficial} prefixes)"
            res.append(f"deliverable to {csirtmail_known} national/governmental CSIRTs {s}{s2}")

        # interesting if using either incident-contact or abusemail
        if ip_local_known or ip_local_unknown:
            l = []
            if ip_local_known:
                l.append(f"{ip_local_known} IP in {prefix_local_known} prefixes")
            if ip_local_unknown:
                l.append(f"unknown contact for {ip_local_unknown} IP in {prefix_local_unknown} prefixes")
            res.append(f"{abusemail_local} e-mail addresses ({', '.join(l)})")

        # interesting if using abusemail
        if ip_abroad_known or ip_abroad_unknown:
            l = []
            if ip_abroad_known:
                l.append(f"{ip_abroad_known} IP in {prefix_abroad_known} prefixes")
            if ip_abroad_unknown:
                l.append(f"unknown contact for {ip_abroad_unknown} IP in {prefix_abroad_unknown} prefixes")
            res.append(f"{abusemail_abroad} abroad e-mail addresses ({', '.join(l)})")

        r = "; ".join(res) + "." if res else ""
        """ XX
        if invalidLines:
            res += "\nThere were {} invalid lines in {} file.".format(invalidLines, self.parser.invalidReg.getPath())"""

        return r

    def source_file_len(self, source_file):
        """ When a source file is reasonably small (100 MB), count the lines by `wc -l`. Otherwise, guess a value.
            XIf we are using stdin instead of a file, determine the value by enlist all the lines.
        """
        size = Path(source_file).stat().st_size
        if size < 100 * 10 ** 6:
            p = subprocess.Popen(['wc', '-l', source_file], stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
            result, err = p.communicate()
            if p.returncode != 0:
                raise IOError(err)
            return int(result.strip().split()[0]), size
        else:
            # bytes / average number of characters on line in sample
            return ceil(size / (len("".join(self.parser.sample)) / len(self.parser.sample)) / 1000000) * 1000000, size

    def start(self):
        q = self.queue

        class queue_stdout(StringIO):
            def write(self, s):
                if q and not q[-1].endswith("\n"):
                    q[-1] += s
                else:
                    q.append(s)
                sys.__stdout__.write(s)
                sys.__stdout__.flush()

        self.stdout = sys.stdout
        sys.stdout = queue_stdout()

        Thread(target=self._stats, daemon=True).start()

    def stop(self):
        if sys.stdout is not self.stdout:
            self.queue.clear()
            sys.stdout = self.stdout
        self.stats_stop.set()

    def pause(self):
        self.stats_stop._flag = 1

    def release(self):
        self.stats_stop._flag = False

    def _stats(self):
        parser = self.parser
        last_count = 0
        speed = 0.1  # speed of refresh
        for tick in count():
            if self.stats_stop._flag is True:
                return
            # do not refresh when stuck (ex: pdb debugging) (as a side effect, clock does not run on the screen)
            if self.stats_stop._flag != 1 and last_count != parser.line_count:
                v = (parser.line_count - last_count) / speed  # current velocity (lines / second) since last time
                if tick < 20:  # in the beginning, refresh quickly; great look & lower performance
                    speed = 0.2
                elif v == 0:
                    speed = 1  # refresh in 1 sec when processing so heavy no line processed since last time
                else:
                    # faster we process, slower we display (to not waste CPU with displaying)
                    # 10^2 lines/s = 1 s, 10^3 ~ 2, 10^4 ~ 3...
                    # But to make the transition more smoothly, make the avg with the last speed (which is weighted 3 times)
                    speed = (log10(v) - 1 + speed * 3) / 4
                    if speed < 0.3:  # but if going too slow, we will not refresh in such a quick interval
                        speed = 0.3

                parser.velocity = round(v) if v > 1 else round(v, 3)
                last_count = parser.line_count

                self.sout_info()
                Whois.quota.check_over()
            sleep(speed)

    def write_statistics(self):
        if Config.get("write_statistics") and not self.parser.stdin:
            # we write statistics.txt only if we're sourcing from a file, not from stdin
            # XX move this to parser.run_analysis and _resolve_again or to Processor – do not rewrite it every time here!
            stat = self.get_stats_phrase()
            if stat:
                Path(self.parser.source_file.parent, "statistics.txt").write_text(stat)



import json
from collections import defaultdict
from urllib import parse

from convey.config import Config
from convey.controller import Controller
from convey.parser import Parser
from convey.types import Types, TypeGroup, Web

Types.refresh()
TypeGroup.init()
Config.integrity_check()
Config.init_verbosity(True, 30)
Config.set("single_query", True)
parser = Parser(prepare=False)
controller = Controller(parser)
unsafe_fields = [Types.code, Types.external]
for target_type in Config.get("webservice_allow_unsafe_fields", section="FIELDS", get=list):
    if hasattr(Types, target_type):
        unsafe_fields.remove(getattr(Types, target_type))


def application(env, start_response):
    """ WSGI launcher. You may expose installed convey as a web service.
        Launch: uwsgi --http :26683 --wsgi-file wsgi.py
        Access: http://localhost:26683/?q=example.com

        XX Since convey internally uses serveral other libraries, publishing web service
        may put the system in a risk. It would be great to implement a method to rather
        whitelist some of the functionalities by default, ex: allow only IP and hostname fields
        that would convert to whois fields only.
    """
    headers = [('Access-Control-Allow-Origin', '*'), ('Content-Type', 'application/json')]
    argument = defaultdict(lambda: None, parse.parse_qsl(env["QUERY_STRING"]))
    if "q" in argument:
        if not argument["q"].strip():
            status = '400 Bad Request'
            response = '{"error": "no input"}'
        else:
            try:
                if "clear" in argument:
                    if argument["clear"] == "web":
                        Web.cache.clear()

                res = parser.set_stdin([argument["q"]]).set_types(argument["type"]).prepare()
                if "field" in argument:
                    target_type = controller.add_new_column(argument["field"], True)
                    if target_type in unsafe_fields:
                        raise SystemExit(f"Unsafe field '{target_type}' disabled via web")
                if res.is_single_query:
                    response = res.run_single_query(json=True)
                    status = '200 OK'
                else:
                    raise SystemExit("Too complicated")
            except SystemExit as e:
                d = {"error": "could not process input"}
                # XX we could change print("..."); quit(); to raising an exception so that we may catch a message
                # Now, everything will pop out to stdout, not to the web browser.
                if str(e):
                    d["message"] = str(e)
                response = json.dumps(d)
                status = '400 Bad Request'
    else:
        status = '400 Bad Request'
        response = '{"error": "put something in parameter \'q\' as query"}'
    start_response(status, headers)

    return [bytes(response, "UTF-8")]



from __future__ import annotations  # remove as of Python3.11 in every file
import io
import logging
import traceback
from bdb import BdbQuit
from collections import defaultdict
from csv import reader as csvreader, writer as csvwriter
from functools import reduce
from operator import eq, ne, mul
from pathlib import Path
from queue import Queue, Empty
from sys import exit
from threading import Thread, Lock
from typing import Dict, TYPE_CHECKING, List, Optional, Tuple, Union

from .action import Expandable
from .attachment import Attachment
from .config import Config
from .dialogue import ask
from .types import Web
from .whois import Quota, UnknownValue

if TYPE_CHECKING:
    import _csv
    from .parser import Parser
    from .definition import Settings


logger = logging.getLogger(__name__)


def prod(iterable):  # XX as of Python3.8, replace with math.prod
    return reduce(mul, iterable, 1)




class Processor:
    """ Opens the CSV file and processes the lines. """

    descriptors: Dict[Union[int, str], Tuple[io.StringIO, _csv._writer]]  # location => file_descriptor, his csv-writer
    descriptorsStatsOpen: Dict[str, int]  # {location: count} XX(performance) we may use a SortedDict object

    def __init__(self, parser, rewrite=True):
        """

        :type rewrite: bool Previously created files will get rewritten by default. Not wanted when we're resolving invalid lines or so.
        """
        self.parser: Parser = parser
        if rewrite:
            parser.files_created.clear()

        self.unique_sets = defaultdict(set)
        self.descriptors_max = 1000  # XX should be given by the system, ex 1024
        self.descriptors_count = 0
        self.descriptorsStatsOpen = {}
        self.descriptorsStatsAll = defaultdict(int)
        self.descriptors = {}
        self._lock = Lock()

    def process_file(self, file, rewrite=False, stdin=None):
        parser = self.parser
        inf = self.parser.informer
        self.__init__(parser, rewrite=rewrite)
        settings = parser.settings.copy()

        # apply setup
        # convert settings["add"] to lambdas
        # Ex: ... = [("netname", 20, [lambda x, lambda x...]), ...]
        settings["addByMethod"] = [(f.name, f.source_field.col_i_original, f.get_methods()) for f in settings["add"]]
        del settings["add"]

        # convert settings["merge"] to lambdas
        if settings["merge"]:
            settings["merging"] = True
            settings["addByMethod"] += [("merged", op.local_column.col_i_original, (lambda x: op.get(x),)) for op in settings["merge"]]
            del settings["merge"]

        # convert filter settings to pre (before line processing) and post that spare a lot of time
        #   (ex: skip lines before WHOIS processing)
        settings["f_pre"] = []
        settings["f_post"] = []
        for it in settings["filter"]:
            col_i = it[1]
            settings["f_post" if parser.fields[col_i].is_new else "f_pre"].append(it)
        del settings["filter"]

        settings["u_pre"] = []
        settings["u_post"] = []
        for col_i in settings["unique"]:  # list of uniqued columns [2, 3, 5, ...]
            settings["u_post" if parser.fields[col_i].is_new else "u_pre"].append(col_i)
        del settings["unique"]

        if [f for f in self.parser.fields if (not f.is_chosen or f.col_i_original != f.col_i)]:
            settings["chosen_cols"] = [f.col_i_original for f in self.parser.fields if f.is_chosen]

        Web.init([f.type for f in self.parser.get_computed_fields()])

        # start file processing
        try:
            if stdin:
                source_stream = stdin
                settings["target_file"] = 1
            else:
                source_stream = open(file, "r")
                settings["target_file"] = str(parser.target_file)
            if parser.stdout is not None:
                settings["target_file"] = 1
            if parser.external_stdout:
                settings["target_file"] = 2
            # with open(file, "r") as sourceF:
            reader = csvreader(source_stream, skipinitialspace=parser.is_pandoc, dialect=parser.dialect)
            if parser.has_header:  # skip header
                reader.__next__()
                if parser.is_pandoc:  # second line is just first line underlining
                    reader.__next__()

            # prepare thread processing
            threads = []
            t = Config.get("threads")
            if t == "auto":
                # XX since threads are experimental, make this disable threads
                # In the future, enable it when using DNS and disable when using WHOIS (that may lead to duplicite calls)
                thread_count = 0
            elif type(t) is bool:
                thread_count = 10 if t else 0
            else:
                thread_count = int(t)

            if thread_count:
                q = Queue(maxsize=thread_count + 2)

                def x():
                    while True:
                        m = q.get()
                        if m is False:  # kill signal received
                            q.task_done()
                            return
                        parser.line_count += 1
                        self.process_line(parser, m, settings)
                        q.task_done()

                for index in range(thread_count):
                    t = Thread(target=x, daemon=True)
                    # prepend zeroes; 10 threads ~ 01..10, 100 threads ~ 001..100
                    t.name = "0" * (len(str(thread_count)) - len(str(index + 1))) + str(index + 1)
                    t.start()
                    threads.append(t)

            inf.start()

            for row in reader:
                try:
                    if not row:  # skip blank
                        continue
                    if thread_count:
                        q.put(row)
                    else:
                        parser.line_count += 1
                        self.process_line(parser, row, settings)
                except BdbQuit:
                    print("Interrupted.")
                    raise
                except KeyboardInterrupt:
                    inf.pause()
                    print(f"Keyboard interrupting on line number: {parser.line_count}")
                    s = "[a]utoskip LACNIC encounters" \
                        if self.parser.queued_lines_count and not Config.get("lacnic_quota_skip_lines", "FIELDS") else ""
                    o = ask("Keyboard interrupt caught. Options: continue (default, do the line again), "
                            "[s]kip the line, [d]ebug, [e]nd processing earlier, [q]uit: ")
                    inf.release()
                    if o == "a":
                        Config.set("lacnic_quota_skip_lines", True, "FIELDS")
                    if o == "d":
                        # I dont know why.
                        print("Maybe you should hit n multiple times because pdb takes you to the wrong scope.")
                        Config.get_debugger().set_trace()
                    elif o == "s":
                        continue  # skip to the next line XXmaybe this is not true for threads
                    elif o == "e":  # end processing now
                        if thread_count:
                            try:
                                while q.get_nowait():  # empty remaining queue
                                    pass
                            except Empty:
                                pass
                            [q.put(False) for _ in threads]  # send kill signal
                            [t.join() for t in threads]  # wait the threads to write descriptors
                        return
                    elif o == "q":
                        self._close_descriptors()
                        exit()
                    else:  # continue from last row
                        parser.line_count -= 1  # let's pretend we didn't just do this row before and give it a second chance
                        if thread_count:
                            # XX when in threads, task_done was not called to queue so no need to requeuing,
                            #  Instead, the line is put to invalid rows. At least when I was digging a number of DNS queries
                            #  that received.
                            # subprocess.CalledProcessError: Command '['dig', '+short', '-t',
                            #           'A', '...', '+timeout=1']' died with <Signals.SIGINT: 2>.
                            # This behaviour was not observed without threads, instead, a KeyboardInterrupt was produced.
                            continue
                        self.process_line(parser, row, settings)

            # join threads
            while True:
                try:
                    [q.put(False) for _ in threads]  # send kill signal
                    [t.join() for t in threads]  # wait the threads to write descriptors
                    break
                except KeyboardInterrupt:
                    print("Cannot keyboard interrupt now, all events were sent to threads.")

            # after processing changes
            if settings["aggregate"]:  # write aggregation results now because we skipped it in process_line when aggregating
                for location, data in self.parser.aggregation.items():
                    if location == 2:  # this is a sign that we store raw data to stdout (not through a CSVWriter)
                        t = parser.external_stdout  # custom object simulate CSVWriter - it adopts .writerow and .close methods
                    elif location == 1:  # this is a sign we output csv data to stdout
                        t = io.StringIO()
                        self.descriptors[1] = t, None  # put the aggregated data to the place we would attend them
                    else:
                        t = open(Path(Config.get_cache_dir(), location), "w")
                        if len(self.parser.aggregation) > 1:
                            print("\nSplit location: " + location)
                    header, rows = inf.get_aggregation(data, nice=False)
                    w = csvwriter(t, dialect=settings["dialect"])
                    w.writerow(header)
                    [w.writerow(r) for r in rows]
        finally:
            inf.stop()
            if not stdin:
                source_stream.close()
            if not self.parser.is_split:
                if 1 in self.descriptors:  # we have all data in a io.TextBuffer, not in a regular file
                    # the data are in the self.descriptors[1]
                    if Config.verbosity <= logging.INFO:
                        print("\n\n** Completed! **\n")
                    result = self.descriptors[1][0].getvalue()
                    print(result.strip())
                    parser.stdout = result
                    parser.stdout_sample = [row.split(parser.dialect.delimiter) for row in result.split("\n", 10)[:10]]
                    self.parser.saved_to_disk = False
                else:
                    self.parser.saved_to_disk = True
            if 2 in self.descriptors:
                # the data were in parser.stdout and were taken by the remote application
                del self.descriptors[2]

            self._close_descriptors()

        if self.parser.is_split:
            # set that a mail with this attachment have not yet been sent
            existing = {a.filename for a in self.parser.attachments}
            self.parser.attachments.extend(Attachment(f) for f in parser.files_created
                                           if f not in existing  # not existing yet
                                           and f not in (Config.INVALID_NAME, Config.UNKNOWN_NAME, Config.QUEUED_NAME))

            if self.parser.is_split and self.parser.stdout is True:
                # Even though the --output True flag was on, nothing was printed out
                # as the contents were saved to the files while being split.
                # Print the files to the screen at least ex post.
                # XX --output [file] does nothing while splitting. In parser.prepare_target_file, we set
                #   self.is_split = True
                #   self.target_file = None
                # We may (a) print out a warning when setting target_file to None if Config.get("output") specified,
                # (b) set the target_file nevertheless and join the contents from split files here.
                [[print(x) for x in (f"* Saved to {a.path.name}", "", a.path.read_text())]
                 for a in self.parser.attachments]
        inf.write_statistics()

    def _close_descriptors(self):
        """ Descriptors have to be closed (flushed) """
        for f in self.descriptors.values():
            f[0].close()

    def process_line(self, parser: Parser, line: List, settings: Settings, fields: Union[Tuple, List]=None):
        """
        Parses line – compute fields while adding, perform filters, pick or delete cols, split and write to a file.
        """
        try:
            add = False
            if not fields:
                fields = line.copy()
                if len(fields) is not len(parser.first_line_fields):
                    raise RuntimeWarning(f"Invalid number of line fields ({len(fields)})")
                add = True

            # pre filtering
            # filter (include, col, value), ex: [(True, 23, "passed-value"), (False, 13, "another-value")]
            for include, col_i, val in settings["f_pre"]:
                if (ne if include else eq)(val, fields[col_i]):
                    return False

            # unique columns
            if settings["u_pre"]:
                for u in settings["u_pre"]:  # list of uniqued columns [2, 3, 5, ...]
                    if fields[u] in self.unique_sets[u]:  # skip line
                        return False
                else:  # do not skip line
                    for u in settings["u_pre"]:
                        self.unique_sets[u].add(fields[u])

            # add fields
            if add:
                list_lengths = []
                for _, col_i, lambdas in settings["addByMethod"]:  # [("netname", 20, [lambda x, lambda x...]), ...]
                    val = fields[col_i]
                    for l in lambdas:
                        if isinstance(val, list):
                            # resolve all items, while flattening any list encountered
                            val = [y for x in (l(v) for v in val) for y in (x if type(x) is list else [x])]
                        else:
                            val = l(val)
                    fields.append(val)
                    if isinstance(val, list):
                        list_lengths.append(len(val) or 1)
                if list_lengths:  # duplicate rows because we received lists amongst scalars
                    row_count = prod(list_lengths)
                    for i, f in enumerate(fields):  # 1st col returns 3 values, 2nd 2 values → both should have 3*2 = 6 values
                        if type(f) is not list:
                            fields[i] = [f]
                        # if we received empty list, row is invalid
                        # ex: missing IP of a hostname
                        if not fields[i]:
                            raise RuntimeWarning(
                                f"Column {i + 1} cannot be determined and the row was marked as invalid")
                        fields[i] *= row_count // len(fields[i])
                    it = zip(*fields)
                    fields = it.__next__()  # now we are sure fields has only scalar values
                    for v in it:  # duplicate row because one of lambdas produced a multiple value list
                        self.process_line(parser, line, settings, v)  # new row with scalar values only

            if settings["merging"]:
                fields = tuple(Expandable.flatten(fields))

            # post filtering
            # filter (include, col, value), ex: [(True, 23, "passed-value"), (False, 13, "another-value")]
            for include, col_i, val in settings["f_post"]:
                if (ne if include else eq)(val, fields[col_i]):
                    return False

            # unique columns
            if settings["u_post"]:
                for u in settings["u_post"]:  # list of uniqued columns [2, 3, 5, ...]
                    if fields[u] in self.unique_sets[u]:  # skip line
                        return False
                else:  # do not skip line
                    for u in settings["u_post"]:
                        self.unique_sets[u].add(fields[u])

            # pick or delete columns
            if settings["chosen_cols"]:
                chosen_fields = [fields[i] for i in settings["chosen_cols"]]  # chosen_cols = [3, 9, 12]
            else:
                chosen_fields = fields

            # determine location
            if type(settings["split"]) == int:
                # split ('/' is a forbidden char in linux file names)
                location = fields[settings["split"]].replace("/", "-")
                if not location:
                    parser.unknown_lines_count += 1
                    location = Config.UNKNOWN_NAME
                    chosen_fields = line  # reset to the original line (location will be reprocessed)
            else:
                location = settings["target_file"]

            # aggregation: count column
            if settings["aggregate"]:
                grp = fields[settings["aggregate"].group_by.col_i] if settings["aggregate"].group_by else None
                for i, (fn, col_data) in enumerate(settings["aggregate"].actions):
                    # counters[location file][grouped row][order in aggregation settings] = [sum generator, count]
                    loc = self.parser.aggregation[location]

                    if grp:  # it makes sense to compute total row because we are grouping
                        if None not in loc:
                            loc[None] = []
                        if len(loc[None]) <= i:
                            loc[None].append([fn(), 0])
                            next(loc[None][i][0])
                        if fn.__name__ == "list":
                            loc[None][i][1] = "(all)"  # we do not want to enlist whole table
                        else:
                            loc[None][i][1] = loc[None][i][0].send(fields[col_data.col_i])

                    if grp not in loc:
                        loc[grp] = []
                    if len(loc[grp]) <= i:
                        loc[grp].append([fn(), 0])
                        next(loc[grp][i][0])
                    loc[grp][i][1] = loc[grp][i][0].send(fields[col_data.col_i])
                return  # we will not write anything right know, aggregation results are not ready yet
        except Quota.QuotaExceeded:
            parser.queued_lines_count += 1
            location = Config.QUEUED_NAME
            chosen_fields = line  # reset the original line (location will be reprocessed)
        except UnknownValue:  # `whois_reprocessable_unknown` is True
            # we do want to reprocess such lines at the end no matter single or split processing
            # (When split processing, this behaviour is obligatory,
            # such lines will be marked unknowns above when determining file location)
            parser.unknown_lines_count += 1
            location = Config.UNKNOWN_NAME
            chosen_fields = line  # reset the original line (location will be reprocessed)
        except BdbQuit:  # prevent BdbQuit to be caught by general Exception below
            raise
        except Exception as e:
            if Config.is_debug():
                traceback.print_exc()
                # cannot post_mortem from here, I could not quit (apart os._exit);
                # looped here again there, catching BdbQuit ignored
                Config.get_debugger().set_trace()
            elif isinstance(e, RuntimeWarning):
                logger.warning(f"Invalid line: {e} on line {line}")
            else:
                logger.warning(e, exc_info=True)
            parser.invalid_lines_count += 1
            location = Config.INVALID_NAME
            chosen_fields = line  # reset the original line (location will be reprocessed)

        if not location:
            return
        elif location in parser.files_created:
            method = "a"
        else:
            method = "w"
            # print("File created", location, parser.delimiter.join(chosen_fields))
            parser.files_created.add(location)

        # choose the right file descriptor for saving
        # (we do not close descriptors immediately, if needed we close the one the least used)
        if location not in self.descriptorsStatsOpen:
            if self.descriptors_count >= self.descriptors_max:  # too many descriptors open, we have to close the least used
                key = min(self.descriptorsStatsOpen, key=self.descriptorsStatsOpen.get)
                self.descriptors[key][0].close()
                # print("Closing", key, self.descriptorsStatsOpen[key])
                del self.descriptorsStatsOpen[key]
                self.descriptors_count -= 1
            # print("Opening", location)

            if location == 2:  # this is a sign that we store raw data to stdout (not through a CSVWriter)
                t = w = parser.external_stdout  # custom object simulate CSVWriter - it adopts .writerow and .close methods
            else:
                if location == 1:  # this is a sign we output csv data to stdout
                    t = io.StringIO()
                else:
                    t = open(Path(Config.get_cache_dir(), location), method)
                w = csvwriter(t, dialect=settings["dialect"])
            self.descriptors[location] = t, w
            self.descriptors_count += 1
        self.descriptorsStatsAll[location] += 1
        self.descriptorsStatsOpen[location] = self.descriptorsStatsAll[location]
        f = self.descriptors[location]
        with self._lock:
            if method == "w" and settings["header"]:
                # write original header (stored unchanged in parser.first_line_fields) if line to be reprocessed or modified header
                if location in (Config.INVALID_NAME, Config.UNKNOWN_NAME, Config.QUEUED_NAME):
                    f[1].writerow(parser.first_line_fields)
                else:
                    f[0].write(parser.header)
            f[1].writerow(chosen_fields)
